#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xc592e403

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from pyprover.constants import imp_sym
from pyprover.constants import and_sym
from pyprover.constants import or_sym
from pyprover.constants import forall_sym
from pyprover.constants import exists_sym

# Functions:

def unorderd_eq(list1, list2):
    if len(list1) != len(list2):
        return False
    for x in list1:
        if x not in list2:
            return False
    return True

def quote(expr, in_quantifier=False):
    """Adds parentheses in the presence of Imp, And, Or, ForAll, and Exists."""
    expr_str = str(expr)
    if in_quantifier and (expr_str.startswith(forall_sym) or expr_str.startswith(exists_sym)):
        return expr_str
    elif (imp_sym in expr_str or and_sym in expr_str or or_sym in expr_str or forall_sym in expr_str or exists_sym in expr_str):
        return "(" + expr_str + ")"
    else:
        return expr_str

def log_simplification(old, new, debug=False, **kwargs):
    """If there was simplification done, displays a message."""
    if debug:
        old_str = str(old)
        new_str = str(new)
        if old_str != new_str:
            print(old_str, "=>", new_str)

def rem_var(var, subs):
    """Deletes any substitutions of var in subs."""
    newsubs = subs.copy()
    try:
        del newsubs[var.constant()]
    except KeyError:
        pass
    try:
        del newsubs[var.variable()]
    except KeyError:
        pass
    return newsubs

class counter(_coconut.object):
    def __init__(self, num=-1):
        self.num = num
    def inc(self):
        assert not self.done(), self.num
        self.num -= 1
    def done(self):
        return self.num == 0

_coconut_call_set_names(counter)
@_coconut_tco
def sub_once(obj, subs):
    """Performs one substitution of subs into obj."""
    return _coconut_tail_call(obj.substitute, subs, counter=counter(1))

def can_sub(kwargs):
    """Determines if the counter in kwargs allows for another sub."""
    try:
        return not kwargs["counter"].done()
    except KeyError:
        return True

def do_sub(kwargs):
    """Increments the counter in kwargs."""
    try:
        kwargs["counter"].inc()
    except KeyError:
        pass

def merge_dicts(dict1, dict2):
    """Merge dictionaries if there are no conflicts, otherwise None."""
    out = {}
    for key, val in _coconut.itertools.chain.from_iterable(_coconut_reiterable(_coconut_func() for _coconut_func in (lambda: dict1.items(), lambda: dict2.items()))):
        if key not in out:
            out[key] = val
        elif out[key] != val:
            return None
    return out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x774454dc

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from pyprover.logic import Proposition
from pyprover.logic import Constant
from pyprover.logic import Implies
from pyprover.logic import wff
from pyprover.logic import bot

# Functions:

@_coconut_tco
def props(names):
    """Constructs propositions from a space-seperated string of names."""
    return _coconut_tail_call(map, Proposition, names.split())

@_coconut_tco
def terms(names):
    """Constructs constants from a space-seperated string of names."""
    return _coconut_tail_call(map, Constant, names.split())

@_coconut_tco
def solve(expr, **kwargs):
    """Converts to CNF and performs all possible resolutions."""
    return _coconut_tail_call(expr.simplify(dnf=False, **kwargs).resolve, **kwargs)
strict_solve = _coconut.functools.partial(solve, nonempty_universe=False)

def no_proof_of(givens, conclusion):
    """Finds a formula that represents the givens not implying the conclusion."""
    if wff(givens):
        givens = (givens,)
    else:
        givens = tuple(givens)
    return ~Implies(*(givens + (conclusion,)))

def proves(givens, conclusion, **kwargs):
    """Determines if the givens prove the conclusion."""
    return (solve)(no_proof_of(givens, conclusion), **kwargs) == bot
strict_proves = _coconut.functools.partial(proves, nonempty_universe=False)

def iff(a, b):
    """Creates a formula for a implies b and b implies a."""
    assert wff(a), a
    assert wff(b), b
    return a >> b & b >> a

def proves_and_proved_by(a, b, **kwargs):
    """Determines if a is true if and only if b."""
    a = a.simplify(dnf=False, **kwargs)
    b = b.simplify(dnf=False, **kwargs)
    return (_coconut.functools.partial(proves, **kwargs))(a, b) and (_coconut.functools.partial(proves, **kwargs))(b, a)
strict_proves_and_proved_by = _coconut.functools.partial(proves_and_proved_by, nonempty_universe=False)

@_coconut_tco
def simplify(expr, *exprs, **kwargs):
    """Simplify the given expression[s]."""
    if exprs:
        return _coconut_tail_call((tuple), (map)(lambda x: x.simplify(**kwargs), (expr,) + exprs))
    else:
        return _coconut_tail_call(expr.simplify, **kwargs)
strict_simplify = _coconut.functools.partial(simplify, nonempty_universe=False)

def simplest_form(expr, **kwargs):
    """Finds the shortest simplification for the given expression."""
    cnf_expr = expr.simplify(dnf=False, **kwargs)
    dnf_expr = cnf_expr.simplify(dnf=True, **kwargs)
    if len(cnf_expr) <= len(dnf_expr):
        return cnf_expr
    else:
        return dnf_expr
strict_simplest_form = _coconut.functools.partial(simplest_form, nonempty_universe=False)

@_coconut_tco
def simplest_solution(expr, **kwargs):
    """Finds the shortest resolved simplification for the given expression."""
    return _coconut_tail_call((simplest_form), (solve)(expr, **kwargs), **kwargs)
strict_simplest_solution = _coconut.functools.partial(simplest_solution, nonempty_universe=False)

@_coconut_tco
def substitute(expr, subs, **kwargs):
    """Substitutes expressions or booleans into the given expression."""
    return _coconut_tail_call(expr.substitute, subs, **kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x2d972424

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from contextlib import contextmanager

from pyprover.tools import props
from pyprover.tools import terms

# Base Class:

class Vars(_coconut.object):
    @classmethod
    def items(cls):
        for name, var in vars(cls).items():
            if not name.startswith("_"):
                yield name, var
    @classmethod
    def use(cls, globs=None):
        """Put variables into the global namespace."""
        if globs is None:
            globs = globals()
        for name, var in cls.items():
            globs[name] = var
    @classmethod
    @contextmanager
    def using(cls, globs=None):
        """Temporarilty put variables into the global namespace."""
        if globs is None:
            globs = globals()
        prevars = {}
        for name, var in cls.items():
            if name in globs:
                prevars[name] = globs[name]
            globs[name] = var
        try:
            yield
        finally:
            for name, var in cls.items():
                if name in prevars:
                    globs[name] = prevars[name]
                else:
                    del globs[name]
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


# Derived Classes:

_coconut_call_set_names(Vars)
class LowercasePropositions(Vars):
    a, b, c = props("a b c")
    d, e, f = props("d e f")
    g, h, i = props("g h i")
    j, k, l = props("j k l")
    m, n, o = props("m n o")
    p, q, r = props("p q r")
    s, t, u = props("s t u")
    v, w, x = props("v w x")
    y, z = props("y z")

_coconut_call_set_names(LowercasePropositions)
class UppercasePropositions(Vars):
    A, B, C = props("A B C")
    D, E, F = props("D E F")
    G, H, I = props("G H I")
    J, K, L = props("J K L")
    M, N, O = props("M N O")
    P, Q, R = props("P Q R")
    S, T, U = props("S T U")
    V, W, X = props("V W X")
    Y, Z = props("Y Z")

_coconut_call_set_names(UppercasePropositions)
class LowercaseVariables(Vars):
    a, b, c = terms("a b c")
    d, e, f = terms("d e f")
    g, h, i = terms("g h i")
    j, k, l = terms("j k l")
    m, n, o = terms("m n o")
    p, q, r = terms("p q r")
    s, t, u = terms("s t u")
    v, w, x = terms("v w x")
    y, z = terms("y z")

_coconut_call_set_names(LowercaseVariables)
class UppercaseVariables(Vars):
    A, B, C = terms("A B C")
    D, E, F = terms("D E F")
    G, H, I = terms("G H I")
    J, K, L = terms("J K L")
    M, N, O = terms("M N O")
    P, Q, R = terms("P Q R")
    S, T, U = terms("S T U")
    V, W, X = terms("V W X")
    Y, Z = terms("Y Z")

_coconut_call_set_names(UppercaseVariables)
class StandardMath(Vars): pass
_coconut_call_set_names(StandardMath)
for name, var in _coconut.itertools.chain.from_iterable(_coconut_reiterable(_coconut_func() for _coconut_func in (lambda: LowercaseVariables.items(), lambda: UppercasePropositions.items()))):
    setattr(StandardMath, name, var)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x216e9829

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from pyparsing import Literal
from pyparsing import ZeroOrMore
from pyparsing import OneOrMore
from pyparsing import Optional
from pyparsing import oneOf
from pyparsing import replaceWith
from pyparsing import Regex
from pyparsing import Forward
from pyparsing import ParserElement
from pyparsing import stringStart
from pyparsing import stringEnd

from pyprover.constants import all_top_syms
from pyprover.constants import all_bot_syms
from pyprover.constants import all_not_syms
from pyprover.constants import all_imp_syms
from pyprover.constants import all_and_syms
from pyprover.constants import all_or_syms
from pyprover.constants import all_forall_syms
from pyprover.constants import all_exists_syms
from pyprover.logic import top
from pyprover.logic import bot
from pyprover.logic import Proposition
from pyprover.logic import Predicate
from pyprover.logic import Constant
from pyprover.logic import Function
from pyprover.logic import Not
from pyprover.logic import Implies
from pyprover.logic import And
from pyprover.logic import Or
from pyprover.logic import Exists
from pyprover.logic import ForAll
from pyprover.logic import Eq

ParserElement.enablePackrat()

oneOf = _coconut_forward_compose(list, oneOf)

# Utilities:

@_coconut_tco
def attach(action, item):
    """Attaches a parse action to an item."""
    return _coconut_tail_call(item.copy().addParseAction, action)

@_coconut_tco
def call(action, item):
    """Call an action on the tokens in item."""
    @_coconut_tco
    def parse_action(o, l, tokens):
        return _coconut_tail_call(action, *tokens)
    return _coconut_tail_call((attach), parse_action, item)

@_coconut_tco
def fixto(output, item):
    """Forces an item to result in a specific output."""
    return _coconut_tail_call((attach), replaceWith(output), item)

def tokenlist(sep, item):
    """Creates a list of tokens matching the item."""
    return item + ZeroOrMore(sep + item) + Optional(sep)

@_coconut_tco
def parse(grammar, text):
    """Parses text using grammar."""
    return _coconut_tail_call(grammar.parseWithTabs().parseString, text)

# Grammar:

@_coconut_tco
@_coconut_mark_as_match
def exists_handle(*_coconut_match_args, **_coconut_match_kwargs):
    _coconut_match_check_0 = False
    _coconut_FunctionMatchError = _coconut_get_function_match_error()
    if (_coconut.len(_coconut_match_args) <= 2) and (_coconut.sum((_coconut.len(_coconut_match_args) > 0, "const" in _coconut_match_kwargs)) == 1) and (_coconut.sum((_coconut.len(_coconut_match_args) > 1, "expr" in _coconut_match_kwargs)) == 1):
        _coconut_match_temp_0 = _coconut_match_args[0] if _coconut.len(_coconut_match_args) > 0 else _coconut_match_kwargs.pop("const")
        _coconut_match_temp_1 = _coconut_match_args[1] if _coconut.len(_coconut_match_args) > 1 else _coconut_match_kwargs.pop("expr")
        if not _coconut_match_kwargs:
            const = _coconut_match_temp_0
            expr = _coconut_match_temp_1
            _coconut_match_check_0 = True
    if not _coconut_match_check_0:
        raise _coconut_FunctionMatchError('match def exists_handle(const, expr) = Exists(const, expr)', _coconut_match_args)

    return _coconut_tail_call(Exists, const, expr)

@addpattern(exists_handle)
@_coconut_tco
@_coconut_mark_as_match
def exists_handle(*_coconut_match_args, **_coconut_match_kwargs):
    _coconut_match_check_1 = False
    _coconut_FunctionMatchError = _coconut_get_function_match_error()
    if (_coconut.len(_coconut_match_args) <= 3) and (_coconut.sum((_coconut.len(_coconut_match_args) > 0, "const" in _coconut_match_kwargs)) == 1) and (_coconut.sum((_coconut.len(_coconut_match_args) > 1, "prop" in _coconut_match_kwargs)) == 1) and (_coconut.sum((_coconut.len(_coconut_match_args) > 2, "expr" in _coconut_match_kwargs)) == 1):
        _coconut_match_temp_0 = _coconut_match_args[0] if _coconut.len(_coconut_match_args) > 0 else _coconut_match_kwargs.pop("const")
        _coconut_match_temp_1 = _coconut_match_args[1] if _coconut.len(_coconut_match_args) > 1 else _coconut_match_kwargs.pop("prop")
        _coconut_match_temp_2 = _coconut_match_args[2] if _coconut.len(_coconut_match_args) > 2 else _coconut_match_kwargs.pop("expr")
        if not _coconut_match_kwargs:
            const = _coconut_match_temp_0
            prop = _coconut_match_temp_1
            expr = _coconut_match_temp_2
            _coconut_match_check_1 = True
    if not _coconut_match_check_1:
        raise _coconut_FunctionMatchError('match def exists_handle(const, prop, expr) =', _coconut_match_args)

    return _coconut_tail_call(Exists, const, prop(const) & expr)

@_coconut_tco
@_coconut_mark_as_match
def forall_handle(*_coconut_match_args, **_coconut_match_kwargs):
    _coconut_match_check_2 = False
    _coconut_FunctionMatchError = _coconut_get_function_match_error()
    if (_coconut.len(_coconut_match_args) <= 2) and (_coconut.sum((_coconut.len(_coconut_match_args) > 0, "const" in _coconut_match_kwargs)) == 1) and (_coconut.sum((_coconut.len(_coconut_match_args) > 1, "expr" in _coconut_match_kwargs)) == 1):
        _coconut_match_temp_0 = _coconut_match_args[0] if _coconut.len(_coconut_match_args) > 0 else _coconut_match_kwargs.pop("const")
        _coconut_match_temp_1 = _coconut_match_args[1] if _coconut.len(_coconut_match_args) > 1 else _coconut_match_kwargs.pop("expr")
        if not _coconut_match_kwargs:
            const = _coconut_match_temp_0
            expr = _coconut_match_temp_1
            _coconut_match_check_2 = True
    if not _coconut_match_check_2:
        raise _coconut_FunctionMatchError('match def forall_handle(const, expr) = ForAll(const, expr)', _coconut_match_args)

    return _coconut_tail_call(ForAll, const, expr)

@addpattern(forall_handle)
@_coconut_tco
@_coconut_mark_as_match
def forall_handle(*_coconut_match_args, **_coconut_match_kwargs):
    _coconut_match_check_3 = False
    _coconut_FunctionMatchError = _coconut_get_function_match_error()
    if (_coconut.len(_coconut_match_args) <= 3) and (_coconut.sum((_coconut.len(_coconut_match_args) > 0, "const" in _coconut_match_kwargs)) == 1) and (_coconut.sum((_coconut.len(_coconut_match_args) > 1, "prop" in _coconut_match_kwargs)) == 1) and (_coconut.sum((_coconut.len(_coconut_match_args) > 2, "expr" in _coconut_match_kwargs)) == 1):
        _coconut_match_temp_0 = _coconut_match_args[0] if _coconut.len(_coconut_match_args) > 0 else _coconut_match_kwargs.pop("const")
        _coconut_match_temp_1 = _coconut_match_args[1] if _coconut.len(_coconut_match_args) > 1 else _coconut_match_kwargs.pop("prop")
        _coconut_match_temp_2 = _coconut_match_args[2] if _coconut.len(_coconut_match_args) > 2 else _coconut_match_kwargs.pop("expr")
        if not _coconut_match_kwargs:
            const = _coconut_match_temp_0
            prop = _coconut_match_temp_1
            expr = _coconut_match_temp_2
            _coconut_match_check_3 = True
    if not _coconut_match_check_3:
        raise _coconut_FunctionMatchError('match def forall_handle(const, prop, expr) =', _coconut_match_args)

    return _coconut_tail_call(ForAll, const, prop(const) >> expr)

class Grammar(_coconut.object):

    lparen = Literal("(").suppress()
    rparen = Literal(")").suppress()
    comma = Literal(",").suppress()
    dot = Literal(".").suppress()
    equals = Literal("=").suppress()
    colon = Literal(":").suppress()

    commalist = _coconut.functools.partial(tokenlist, comma)

    not_op = oneOf(all_not_syms).suppress()
    imp_op = oneOf(all_imp_syms).suppress()
    and_op = oneOf(all_and_syms).suppress()
    or_op = oneOf(all_or_syms).suppress()
    forall_op = oneOf(all_forall_syms).suppress()
    exists_op = oneOf(all_exists_syms).suppress()

    top_lit = (fixto)(top, oneOf(all_top_syms))
    bot_lit = (fixto)(bot, oneOf(all_bot_syms))

    lowercase_name = Regex("[a-z0-9_]\w*")
    uppercase_name = Regex("[A-Z]\w*")

    var = (call)(Constant, lowercase_name)
    func = Forward()
    term = func | var
    terms = lparen - commalist(term) - rparen
    func <<= (call)(Function, lowercase_name + terms)

    prop = (call)(Proposition, uppercase_name)
    pred = (call)(Predicate, uppercase_name + terms)
    equality = (call)(Eq, term + equals + term)
    atom = pred | prop | equality

    expr = Forward()
    quant_sep = dot | comma
    exists = ((call)(exists_handle, exists_op + var + Optional(colon + prop) + quant_sep + expr))
    forall = ((call)(forall_handle, forall_op + var + Optional(colon + prop) + quant_sep + expr))
    quant = exists | forall

    base_expr = top_lit | bot_lit | quant | atom | lparen - expr - rparen
    not_expr = quant | ((call)(Not, not_op + base_expr)) | base_expr | quant
    and_expr = quant | ((call)(And, not_expr + OneOrMore(and_op - not_expr))) | not_expr
    or_expr = quant | ((call)(Or, and_expr + OneOrMore(or_op - and_expr))) | and_expr
    expr <<= quant | ((call)(Implies, or_expr + OneOrMore(imp_op - or_expr))) | or_expr

    formula = stringStart + expr + stringEnd

_coconut_call_set_names(Grammar)
for varname, val in vars(Grammar).items():
    if isinstance(val, ParserElement):
        setattr(Grammar, varname, val.setName(varname))

# Endpoint:

def expr(formula):
    """Parses the given formula into an expression."""
    result = parse(Grammar.formula, formula)
    assert len(result) == 1, results
    return result[0]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0xe1c55f55

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

sys = _coconut_sys

# Utilities:

def first_encodeable(symbols):
    for sym in symbols:
        try:
            if hasattr(sys.stdout, "encoding") and sys.stdout.encoding is not None:
                sym.encode(sys.stdout.encoding)
            else:
                sym.encode()
        except UnicodeEncodeError:
            pass
        else:
            return sym
    raise ValueError("No encodable symbol in " + repr(symbols))

# Installation:

version = "0.5.4"
requirements = ["pyparsing",]
classifiers = ["Development Status :: 3 - Alpha", "License :: OSI Approved :: Apache Software License", "Topic :: Software Development :: Libraries :: Python Modules", "Operating System :: OS Independent",]

# Symbols:

all_top_syms = "top", "\u22a4", "\u252c", "true"
top_sym = (first_encodeable)(all_top_syms)

all_bot_syms = "bot", "\u22a5", "\u2534", "_|_", "false"
bot_sym = (first_encodeable)(all_bot_syms)

all_not_syms = "~", "\xac", "-", "not"
not_sym = (first_encodeable)(all_not_syms)

all_imp_syms = "->", "\u2192"
imp_sym = (first_encodeable)(all_imp_syms)

all_and_syms = "&", "\u2227", "/\\", "and"
and_sym = (first_encodeable)(all_and_syms)

all_or_syms = "|", "\u2228", "\\/", "or"
or_sym = (first_encodeable)(all_or_syms)

all_forall_syms = "FA", "fa", "forall", "A", "\u2200"
forall_sym = (first_encodeable)(all_forall_syms)

all_exists_syms = "TE", "te", "exists", "E", "\u2203"
exists_sym = (first_encodeable)(all_exists_syms)

empty_var = "_"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x67040014

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from pyprover.constants import top_sym
from pyprover.constants import bot_sym
from pyprover.constants import not_sym
from pyprover.constants import imp_sym
from pyprover.constants import and_sym
from pyprover.constants import or_sym
from pyprover.constants import forall_sym
from pyprover.constants import exists_sym
from pyprover.constants import empty_var
from pyprover.util import unorderd_eq
from pyprover.util import quote
from pyprover.util import log_simplification
from pyprover.util import rem_var
from pyprover.util import can_sub
from pyprover.util import do_sub
from pyprover.util import merge_dicts
from pyprover.util import sub_once

# Functions:

def wff(expr):
    """Determines whether expr is a well-formed formula."""
    return isinstance(expr, Expr) and not isinstance(expr, Term)

@_coconut_tco
def isvar(var):
    """Whether a term is a variable."""
    return _coconut_tail_call(isinstance, var, (Const, Var))

# Classes:

class Expr(_coconut.object):
    """Base class for all formulae."""
    __slots__ = ()

    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

    @_coconut_tco
    def __and__(self, other):
        if isinstance(other, And):
            return other & self
        else:
            return _coconut_tail_call(And, self, other)
    @_coconut_tco
    def __or__(self, other):
        if isinstance(other, Or):
            return other | self
        else:
            return _coconut_tail_call(Or, self, other)
    @_coconut_tco
    def __rshift__(self, other):
        if isinstance(other, Imp):
            return other << self
        else:
            return _coconut_tail_call(Imp, self, other)
    def __lshift__(self, other):
        assert wff(other), other
        return other >> self
    @_coconut_tco
    def __invert__(self):
        return _coconut_tail_call(Not, self)
    @_coconut_tco
    def __xor__(self, other):
        return _coconut_tail_call(Or, And(self, Not(other)), And(Not(self), other))
    def __len__(self):
        return 1
    def simplify(self, **kwargs):
        """Simplify the given expression."""
        return self
    def substitute(self, subs, **kwargs):
        """Substitutes a dictionary into the expression."""
        return self
    @_coconut_tco
    def resolve(self, **kwargs):
        """Performs resolution on the clauses in a CNF expression."""
        return _coconut_tail_call(self.simplify, dnf=False, **kwargs)
    @_coconut_tco
    def find_unification(self, other):
        """Find a substitution in self that would make self into other."""
        if isinstance(other, Quantifier):
            return _coconut_tail_call(other.find_unification, self)
        elif self == other:
            return {}
        else:
            return None
    @_coconut_tco
    def contradicts(self, other, **kwargs):
        """Assuming self is simplified, determines if it contradicts other."""
        if isinstance(other, Not):
            return _coconut_tail_call(other.contradicts, self, **kwargs)
        else:
            return self == Not(other).simplify(**kwargs)
    @_coconut_tco
    def resolve_against(self, other, **kwargs):
        """Attempt to perform a resolution against other else None."""
        if isinstance(other, (Not, Or, Eq)):
            return _coconut_tail_call(other.resolve_against, self, **kwargs)
        elif (self.find_unification)(Not(other).simplify(**kwargs)) is not None:
            return bot
        else:
            return None
    def admits_empty_universe(self):
        """Determines if self allows for the possibility of an empty universe."""
        return True

_coconut_call_set_names(Expr)
class Top(Expr):
    """True"""
    __slots__ = ()
    @_coconut_tco
    def __eq__(self, other):
        return _coconut_tail_call(isinstance, other, Top)
    def __repr__(self):
        return "top"
    def __str__(self):
        return top_sym
    def __bool__(self):
        return True
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Top)
top = true = Top()

class Bot(Expr):
    """False"""
    __slots__ = ()
    @_coconut_tco
    def __eq__(self, other):
        return _coconut_tail_call(isinstance, other, Bot)
    def __repr__(self):
        return "bot"
    def __str__(self):
        return bot_sym
    def __bool__(self):
        return False
    def admits_empty_universe(self):
        return False
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Bot)
bot = false = Bot()

class Atom(Expr):
    """Base class for all variables."""
    __slots__ = ("name",)
    def __init__(self, name):
        if isinstance(name, Atom):
            name = name.name
        assert isinstance(name, str), name
        self.name = name
    def __repr__(self):
        return self.__class__.__name__ + '("' + self.name + '")'
    def __str__(self):
        return self.name
    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name
    @_coconut_tco
    def __hash__(self):
        return _coconut_tail_call((hash), (self.__class__.__name__, self.name))
    def substitute_elements(self, subs, **kwargs):
        """Substitute for the elements of the Atom, not the Atom itself."""
        return self
    @_coconut_tco
    def substitute(self, subs, **kwargs):
        if not can_sub(kwargs):
            return self
        _coconut_match_to_0 = subs
        _coconut_match_check_0 = False
        if _coconut.isinstance(_coconut_match_to_0, _coconut.abc.Mapping):
            _coconut_match_temp_0 = _coconut_match_to_0.get(self, _coconut_sentinel)
            if _coconut_match_temp_0 is not _coconut_sentinel:
                sub = _coconut_match_temp_0
                _coconut_match_check_0 = True
        if _coconut_match_check_0:
            do_sub(kwargs)
            if wff(sub):
                return sub
            elif sub is True:
                return top
            elif sub is False:
                return bot
            else:
                raise TypeError("cannot perform substitution " + str(self) + " => " + str(sub))
        else:
            return _coconut_tail_call(self.substitute_elements, subs, **kwargs)

    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(Atom)
class Prop(Atom):
    """Logical proposition that is either true or false."""
    __slots__ = ()
    @_coconut_tco
    def __call__(self, *args):
        return _coconut_tail_call(Pred, self.name, *args)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Prop)
Proposition = Prop

class FuncAtom(Atom):
    """Base class for predicates and functions."""
    __slots__ = ("args",)
    def __init__(self, name, *args):
        super(FuncAtom, self).__init__(name)
        for arg in args:
            assert isinstance(arg, Term), arg
        self.args = args
    def __repr__(self):
        return self.__class__.__name__ + '("' + self.name + '"' + (", " if self.args else "") + ", ".join((repr(x) for x in self.args)) + ")"
    def __str__(self):
        return self.name + "(" + ", ".join((str(x) for x in self.args)) + ")"
    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name and self.args == other.args
    @_coconut_tco
    def __hash__(self):
        return _coconut_tail_call((hash), (self.__class__.__name__, self.name, self.args))
    @_coconut_tco
    def find_unification(self, other):
        if isinstance(other, Quantifier):
            return _coconut_tail_call(other.find_unification, self)
        elif isinstance(other, self.__class__) and self.name == other.name and len(self.args) == len(other.args):
            subs = {}
            for i, x in enumerate(self.args):
                y = other.args[i]
                unif = x.find_unification(y)
                if unif is None:
                    return None
                for var, sub in unif.items():
                    if var not in subs:
                        subs[var] = sub
                    elif subs[var] != sub:
                        return None
            return subs
        else:
            return None
    @_coconut_tco
    def admits_empty_universe(self):
        return _coconut_tail_call(all, (x.admits_empty_universe() for x in self.args))
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(FuncAtom)
class Pred(FuncAtom):
    """Boolean function of terms."""
    __slots__ = ()
    @_coconut_tco
    def proposition(self):
        return _coconut_tail_call(Prop, self.name)
    @_coconut_tco
    def substitute_elements(self, subs, **kwargs):
        if not can_sub(kwargs):
            return self
        _coconut_match_to_1 = subs
        _coconut_match_check_1 = False
        if _coconut.isinstance(_coconut_match_to_1, _coconut.abc.Mapping):
            _coconut_match_temp_0 = _coconut_match_to_1.get(self.proposition(), _coconut_sentinel)
            if _coconut_match_temp_0 is not _coconut_sentinel:
                sub = _coconut_match_temp_0
                _coconut_match_check_1 = True
        if _coconut_match_check_1:
            assert isinstance(sub, Atom), sub
            do_sub(kwargs)
            name = sub.name
        else:
            name = self.name
        if can_sub(kwargs):
            return (Pred)(name, *(map)(_coconut.operator.methodcaller("substitute", subs, **kwargs), self.args))
        else:
            return _coconut_tail_call(Pred, name, *self.args)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Pred)
Predicate = Pred

class Term(Atom):
    """Base class for all terms."""
    __slots__ = ()
    @_coconut_tco
    def variable(self):
        """Convert to a variable."""
        return _coconut_tail_call(Var, self.name)
    @_coconut_tco
    def constant(self):
        """Convert to a constant."""
        return _coconut_tail_call(Const, self.name)
    @_coconut_tco
    def rename(self, name):
        """Create a new term with a different name."""
        return _coconut_tail_call(self.__class__, name)
    @_coconut_tco
    def prime(self):
        """Rename by adding a prime."""
        return _coconut_tail_call(self.rename, self.name + "'")
    @_coconut_tco
    def substitute(self, subs, **kwargs):
        if can_sub(kwargs):
            for var, sub in subs.items():
                if can_sub(kwargs) and isinstance(var, Term) and self.name == var.name:
                    do_sub(kwargs)
                    if isvar(self) or self == var:
                        return sub
                    else:
                        return self.rename(sub.name)
        if can_sub(kwargs):
            return _coconut_tail_call(self.substitute_elements, subs, **kwargs)
        return self
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(Term)
class Var(Term):
    """A variable quantified by a ForAll."""
    __slots__ = ()
    def variable(self):
        return self
    @_coconut_tco
    def __call__(self, *args):
        return _coconut_tail_call(Func, self.name, *args)
    def find_unification(self, other):
        if isinstance(other, Term):
            return {self: other}
        else:
            return None
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Var)
Variable = Var

class Const(Term):
    """A variable quantified by an Exists."""
    __slots__ = ()
    def constant(self):
        return self
    @_coconut_tco
    def __call__(self, *args):
        return _coconut_tail_call(Func, self.name, *args)
    @_coconut_tco
    def find_unification(self, other):
        if isinstance(other, Var):
            return {other: self}
        else:
            return _coconut_tail_call(super(Const, self).find_unification, other)
    def admits_empty_universe(self):
        return False
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Const)
Constant = Const

class Func(Term, FuncAtom):
    """A function on terms."""
    __slots__ = ()
    def substitute_elements(self, subs, **kwargs):
        return (Func)(self.name, *(map)(_coconut.operator.methodcaller("substitute", subs, **kwargs), self.args))
    @_coconut_tco
    def rename(self, name):
        return _coconut_tail_call(self.__class__, name, *self.args)
    @_coconut_tco
    def find_unification(self, other):
        if isinstance(other, Var):
            return {other: self}
        else:
            return _coconut_tail_call(super(Func, self).find_unification, other)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Func)
Function = Func

class UnaryOp(Expr):
    """Base class for unary operators."""
    __slots__ = ("elem",)
    def __init__(self, elem):
        assert wff(elem), elem
        self.elem = elem
    def __repr__(self):
        return self.__class__.__name__ + "(" + repr(self.elem) + ")"
    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.elem == other.elem
    def __str__(self):
        return self.opstr + quote(self.elem)
    def __len__(self):
        return len(self.elem) + 1
    @_coconut_tco
    def substitute(self, subs, **kwargs):
        return _coconut_tail_call(self.__class__, self.elem.substitute(subs, **kwargs))
    @_coconut_tco
    def find_unification(self, other):
        if isinstance(other, Quantifier):
            return _coconut_tail_call(other.find_unification, self)
        elif isinstance(other, self.__class__):
            return _coconut_tail_call(self.elem.find_unification, other.elem)
        else:
            return None
    @_coconut_tco
    def resolve(self, **kwargs):
        return _coconut_tail_call(self.__class__(self.elem.resolve(**kwargs)).simplify, dnf=False, **kwargs)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(UnaryOp)
class Not(UnaryOp):
    """Logical not."""
    __slots__ = ()
    opstr = not_sym
    @property
    def neg(self):
        return self.elem
    @_coconut_tco
    def simplify(self, **kwargs):
        if top == self.neg:
            return bot
        elif bot == self.neg:
            return top
        elif isinstance(self.neg, Not):
            return _coconut_tail_call(self.neg.neg.simplify, **kwargs)
        elif isinstance(self.neg, And):
            return _coconut_tail_call(Or(*map(Not, self.neg.ands)).simplify, **kwargs)
        elif isinstance(self.neg, Or):
            return _coconut_tail_call(And(*map(Not, self.neg.ors)).simplify, **kwargs)
        elif isinstance(self.neg, Imp):
            ands = self.neg.conds + (Not(self.neg.concl),)
            return _coconut_tail_call(And(*ands).simplify, **kwargs)
        elif isinstance(self.neg, Exists):
            return _coconut_tail_call(ForAll, self.neg.var, Not(self.neg.elem).simplify(**kwargs))
        elif isinstance(self.neg, ForAll):
            return _coconut_tail_call(Exists, self.neg.var, Not(self.neg.elem).simplify(**kwargs))
        else:
            return _coconut_tail_call(Not, self.neg.simplify(**kwargs))
    def contradicts(self, other, **kwargs):
        return self.neg == other
    @_coconut_tco
    def resolve_against(self, other, **kwargs):
        if isinstance(other, (Or, Eq)):
            return _coconut_tail_call(other.resolve_against, self, **kwargs)
        elif self.neg.find_unification(other) is not None:
            return bot
        else:
            return None
    @_coconut_tco
    def admits_empty_universe(self):
        if isinstance(self.neg, Atom):
            return _coconut_tail_call(self.neg.admits_empty_universe)
        else:
            return not self.neg.admits_empty_universe()
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(Not)
class Quantifier(Expr):
    """Base class for logical quantifiers."""
    __slots__ = ("var", "elem")
    def __repr__(self):
        return self.__class__.__name__ + '("' + str(self.var) + '", ' + repr(self.elem) + ")"
    def __str__(self):
        return self.opstr + " " + str(self.var) + ", " + quote(self.elem, in_quantifier=True)
    def __len__(self):
        return len(self.elem) + len(self.var)
    @_coconut_tco
    def change_var(self, var):
        """Create an equivalent expression with a new quantified variable."""
        return _coconut_tail_call(self.__class__, var, self.elem.substitute({self.var: var}))
    @_coconut_tco
    def change_elem(self, elem):
        """Create an equivalent quantifier with a new expression."""
        return _coconut_tail_call(self.__class__, self.var, elem)
    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.elem == other.change_var(self.var).elem
        else:
            return False
    def inner_kwargs(self, kwargs):
        inner_kwargs = kwargs.copy()
        inner_kwargs["in_" + self.__class__.__name__.lower()] = True
        return inner_kwargs
    @_coconut_tco
    def simplify(self, **kwargs):
        return _coconut_tail_call(self.__class__(self.var, self.elem.simplify(**self.inner_kwargs(kwargs))).drop_quantifier, **kwargs)
    @_coconut_tco
    def substitute(self, subs, **kwargs):
        return _coconut_tail_call((self.change_elem), (self.elem.substitute)((rem_var)(self.var, subs), **kwargs))
    @_coconut_tco
    def make_free_in(self, other):
        """Makes self free in other."""
        var = self.var
        newvar = var.prime()
        while other != other.substitute({var: newvar}):
            var, newvar = newvar, newvar.prime()
        return _coconut_tail_call(self.change_var, var)
    @_coconut_tco
    def find_unification(self, other):
        unif = self.elem.find_unification(other)
        if unif is None:
            return None
        else:
            return _coconut_tail_call((rem_var), self.var, unif)
    @_coconut_tco
    def resolve_against(self, other, **kwargs):
        if isinstance(other, Quantifier):
            resolution = (self.elem.resolve_against)(Not(other.elem).simplify(**kwargs), **kwargs)
            if resolution is None:
                return None
            elif isinstance(other, ForAll):  # don't pull an Exists out of a ForAll
                return _coconut_tail_call((other.change_elem), (self.change_elem)(resolution))
            else:
                return _coconut_tail_call((self.change_elem), (other.change_elem)(resolution))
        else:
            return _coconut_tail_call(super(Quantifier, self).resolve_against, other, **kwargs)
    @classmethod
    @_coconut_tco
    def blank(cls, elem):
        """Make a quantifier without a variable."""
        return _coconut_tail_call(cls(empty_var, elem).make_free_in, elem)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(Quantifier)
class ForAll(Quantifier):
    """Universal quantifier."""
    __slots__ = ()
    opstr = forall_sym
    def __init__(self, var, elem):
        assert wff(elem), elem
        if isinstance(var, str):
            var = Const(var)
        assert isvar(var), var
        self.var = var.variable()
        self.elem = elem.substitute({var: self.var.variable()})
    @_coconut_tco
    def resolve(self, **kwargs):
        inner_kwargs = self.inner_kwargs(kwargs)
        inner_kwargs["variables"] = kwargs.get("variables", ()) + (self.var,)
        return _coconut_tail_call(ForAll(self.var, self.elem.resolve(**inner_kwargs)).simplify, dnf=False, **kwargs)
    def drop_quantifier(self, nonempty_universe=True, in_forall=False, **kwargs):
        kwargs["nonempty_universe"], kwargs["in_forall"] = nonempty_universe, in_forall
        if not nonempty_universe and not in_forall:
            elem = self.elem
            while isinstance(elem, Exists):
                elem = elem.elem
            if top == elem:
                return elem
        elif self.elem == self.elem.substitute({self.var: self.var.prime()}):
            return self.elem
        return self
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(ForAll)
FA = ForAll

class Exists(Quantifier):
    """Existential quantifier."""
    __slots__ = ()
    opstr = exists_sym
    def __init__(self, var, elem):
        assert wff(elem), elem
        if isinstance(var, str):
            var = Const(var)
        assert isvar(var), var
        self.var = var.constant()
        self.elem = elem.substitute({var: self.var.constant()})
    @_coconut_tco
    def resolve(self, **kwargs):
        inner_kwargs = self.inner_kwargs(kwargs)
        variables = inner_kwargs.get("variables")
        if variables is None:
            skolem_elem = self.elem
        else:
            skolem_var = Func(self.var.name, *variables)
            skolem_elem = self.elem.substitute({self.var: skolem_var})
        return _coconut_tail_call(Exists(self.var, skolem_elem.resolve(**inner_kwargs)).simplify, dnf=False, **kwargs)
    def drop_quantifier(self, nonempty_universe=True, in_exists=False, **kwargs):
        kwargs["nonempty_universe"], kwargs["in_exists"] = nonempty_universe, in_exists
        if not nonempty_universe and not in_exists:
            elem = self.elem
            while isinstance(elem, ForAll):
                elem = elem.elem
            if bot == elem:
                return elem
        elif self.elem == self.elem.substitute({self.var: self.var.prime()}):
            return self.elem
        return self
    def admits_empty_universe(self):
        return False
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Exists)
TE = Exists

class BinaryOp(Expr):
    """Base class for binary operators."""
    __slots__ = ("elems",)
    identity = None
    @_coconut_tco
    def __new__(cls, *elems):
        if not elems:
            if cls.identity is None:
                raise TypeError(cls.__name__ + " requires at least one argument")
            else:
                return cls.identity
        elif len(elems) == 1:
            assert wff(elems[0]), elems[0]
            return elems[0]  # sometimes returns an instance of cls
        else:
            return _coconut_tail_call(super(BinaryOp, cls).__new__, cls)
    def __init__(self, *elems):
        if len(elems) > 1:  # __new__ should handle all other cases
            assert len(elems) >= 2, elems
            for x in elems:
                assert wff(x), x
            self.elems = elems
    def __repr__(self):
        return self.__class__.__name__ + "(" + ", ".join((repr(x) for x in self.elems)) + ")"
    @_coconut_tco
    def __str__(self):
        return _coconut_tail_call((" " + self.opstr + " ").join, (quote(x) for x in self.elems))
    def __len__(self):
        return sum(map(len, self.elems)) + 1
    def substitute(self, subs, **kwargs):
        return (self.__class__)(*(map)(_coconut.operator.methodcaller("substitute", subs, **kwargs), self.elems))
    @_coconut_tco
    def resolve(self, **kwargs):
        elems = (map)(_coconut.operator.methodcaller("resolve", **kwargs), self.elems)
        return _coconut_tail_call(self.__class__(*elems).simplify, dnf=False, **kwargs)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(BinaryOp)
class Imp(BinaryOp):
    """Logical implication."""
    __slots__ = ()
    opstr = imp_sym
    @_coconut_tco
    def __rshift__(self, other):
        if isinstance(other, Imp):
            return _coconut_tail_call(Imp, self, *other.elems)
        else:
            return _coconut_tail_call(Imp, self, other)
    @_coconut_tco
    def __lshift__(self, other):
        return _coconut_tail_call((Imp), *(other,) + self.elems)
    @property
    def conds(self):
        return self.elems[:-1]
    @property
    def concl(self):
        return self.elems[-1]
    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.concl == other.concl and (unorderd_eq)(self.conds, other.conds)
    @_coconut_tco
    def to_or(self):
        ors = tuple(map(Not, self.conds)) + (self.concl,)
        return _coconut_tail_call(Or, *ors)
    @_coconut_tco
    def simplify(self, **kwargs):
        return _coconut_tail_call(self.to_or().simplify, **kwargs)
    @_coconut_tco
    def admits_empty_universe(self):
        return _coconut_tail_call(self.to_or().admits_empty_universe)
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Imp)
Implies = Imp

class BoolOp(BinaryOp):
    """Base class for Or and And."""
    __slots__ = ()
    def __eq__(self, other):
        return isinstance(other, self.__class__) and (unorderd_eq)(self.elems, other.elems)
    def simplify(self, **kwargs):
        elems = (map)(_coconut.operator.methodcaller("simplify", **kwargs), self.merge().elems)
        out = self.__class__(*elems).clean()
        if isinstance(out, self.__class__):
            out = out.distribute(**kwargs)
        if isinstance(out, self.__class__):
            out = out.merge().dedupe()
        if isinstance(out, self.__class__):
            out = out.inner_simplify(**kwargs)
        if isinstance(out, self.__class__):
            out = out.prenex(**kwargs)
        log_simplification(self, out, **kwargs)
        return out
    @_coconut_tco
    def merge(self):
        """Merges nested copies of a boolean operator."""
        elems = []
        for x in self.elems:
            if isinstance(x, self.__class__):
                elems.extend(x.merge().elems)
            else:
                elems.append(x)
        return _coconut_tail_call(self.__class__, *elems)
    @_coconut_tco
    def dedupe(self):
        """Removes duplicate elements from a boolean operator."""
        elems = []
        for x in self.elems:
            if x not in elems:
                elems.append(x)
        return _coconut_tail_call(self.__class__, *elems)
    @_coconut_tco
    def clean(self):
        """Removes copies of the identity."""
        return _coconut_tail_call((self.__class__), *(filter)(_coconut.functools.partial(_coconut.operator.ne, self.identity), self.elems))
    def prenex(self, **kwargs):
        """Pulls quantifiers out."""
        for i, x in enumerate(self.elems):
            if isinstance(x, Quantifier) and self.can_prenex(x, **kwargs):
                elems = self.elems[:i] + self.elems[i + 1:]
                free_x = x.make_free_in(self.__class__(*elems))
                elems += (free_x.elem,)
                return free_x.change_elem(self.__class__(*elems)).simplify(**kwargs)
        return self
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(BoolOp)
class Or(BoolOp):
    """Logical disjunction."""
    __slots__ = ()
    opstr = or_sym
    identity = bot
    @_coconut_tco
    def __or__(self, other):
        return _coconut_tail_call((Or), *self.elems + (other,))
    @property
    def ors(self):
        return self.elems
    def distribute(self, dnf=False, **kwargs):
        """If this Or contains an And, distribute into it."""
        kwargs["dnf"] = dnf
        if not dnf:
            for i, x in enumerate(self.ors):
                if isinstance(x, And):
                    ands = ((Or)(*(y,) + self.ors[:i] + self.ors[i + 1:]) for y in x.ands)
                    return And(*ands).simplify(**kwargs)
        return self
    def inner_simplify(self, nonempty_universe=True, **kwargs):
        """Determines if the Or is a blatant tautology."""
        kwargs["nonempty_universe"] = nonempty_universe
        for i, x in enumerate(self.ors):
            if top == x:
                return x
            for y in self.ors[i + 1:]:
                if x.contradicts(y, **kwargs):
                    if not nonempty_universe and not self.admits_empty_universe():
                        return Exists.blank(top)
                    else:
                        return top
        return self
    def can_prenex(self, other, nonempty_universe=True, in_forall=False, **kwargs):
        kwargs["nonempty_universe"], kwargs["in_forall"] = nonempty_universe, in_forall
        return (nonempty_universe or in_forall or not isinstance(other, Exists) or not self.admits_empty_universe())
    @_coconut_tco
    def resolve_against(self, other, **kwargs):
        if isinstance(other, Eq):
            return _coconut_tail_call(other.resolve_against, self)
        elif isinstance(other, Or):
            not_other_ors = (map)(_coconut.operator.methodcaller("simplify", **kwargs), (map)(Not, other.ors))
            for i, x in enumerate(self.ors):
                if isinstance(x, Eq):
                    resolved_other = (x.paramodulant)(other)
                    return (Or)(*self.ors[:i] + self.ors[i + 1:] + resolved_other.ors)
                for j, y in enumerate(not_other_ors):
                    if isinstance(other.ors[j], Eq):
                        y = other.ors[j]
                        resolved_self = (y.paramodulant)(self)
                        return (Or)(*other.ors[:j] + other.ors[j + 1:] + resolved_self.ors)
                    subs = x.find_unification(y)
                    if subs is not None:
                        return (Or)(*(map)(_coconut.operator.methodcaller("substitute", subs, **kwargs), self.ors[:i] + self.ors[i + 1:] + other.ors[:j] + other.ors[j + 1:]))
        else:
            not_other = Not(other).simplify(**kwargs)
            for i, x in enumerate(self.ors):
                if isinstance(x, Eq):
                    return (x.paramodulant)((Or)(*self.ors[:i] + self.ors[i + 1:]))
                subs = x.find_unification(not_other)
                if subs is not None:
                    return (Or)(*(map)(_coconut.operator.methodcaller("substitute", subs, **kwargs), self.ors[:i] + self.ors[i + 1:]))
        return None
    @_coconut_tco
    def admits_empty_universe(self):
        return _coconut_tail_call(any, (x.admits_empty_universe() for x in self.elems))
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(Or)
class And(BoolOp):
    """Logical conjunction."""
    __slots__ = ()
    opstr = and_sym
    identity = top
    @_coconut_tco
    def __and__(self, other):
        return _coconut_tail_call((And), *self.elems + (other,))
    @property
    def ands(self):
        return self.elems
    def distribute(self, dnf=False, **kwargs):
        """If this And contains an Or, distribute into it."""
        kwargs["dnf"] = dnf
        if dnf:
            for i, x in enumerate(self.ands):
                if isinstance(x, Or):
                    ors = ((And)(*(y,) + self.ands[:i] + self.ands[i + 1:]) for y in x.ors)
                    return Or(*ors).simplify(**kwargs)
        return self
    def inner_simplify(self, nonempty_universe=True, **kwargs):
        """Determines if the And is a blatant contradiction."""
        kwargs["nonempty_universe"] = nonempty_universe
        for i, x in enumerate(self.ands):
            if bot == x:
                return x
            for y in self.ands[i + 1:]:
                if x.contradicts(y, **kwargs):
                    if not nonempty_universe and self.admits_empty_universe():
                        return ForAll.blank(bot)
                    else:
                        return bot
        return self
    def can_prenex(self, other, nonempty_universe=True, in_exists=False, **kwargs):
        kwargs["nonempty_universe"], kwargs["in_exists"] = nonempty_universe, in_exists
        return (nonempty_universe or in_exists or not isinstance(other, ForAll) or all((isinstance(x, ForAll) for x in self.elems)))
    @_coconut_tco
    def resolve(self, nonempty_universe=True, debug=False, **kwargs):
        """Performs all possible resolutions within the And."""
        kwargs["nonempty_universe"], kwargs["debug"] = nonempty_universe, debug
        resolved = super(And, self).resolve(**kwargs)
        if not isinstance(resolved, And):
            log_simplification(self, resolved, **kwargs)
            return resolved
        clauses = (list)(resolved.ands)
        quantifiers = []
        if not nonempty_universe and not self.admits_empty_universe():
            blank = Exists.blank(top)
            (quantifiers.append)(blank.change_elem)
            kwargs = (blank.inner_kwargs)(kwargs)
        prev_clause_len = 1
        while prev_clause_len < len(clauses):
            prev_clause_len = len(clauses)
# reversed ensures conclusions get tested first
            for i in (reversed)(range(1, len(clauses))):
                x = clauses[i]
                for y in clauses[:i + 1]:  # allow resolution of a clause against itself
                    resolution = x.resolve_against(y)
                    if resolution is not None:
                        resolution = resolution.simplify(dnf=False, **kwargs)
                        if debug:
                            print(x, "+", y, "=>", resolution)
                        new_quantifiers = []
                        inner_kwargs = kwargs
                        while isinstance(resolution, Quantifier) and self.can_prenex(resolution, **kwargs):
                            (new_quantifiers.append)(resolution.change_elem)
                            inner_kwargs = (resolution.inner_kwargs)(inner_kwargs)
                            resolution = resolution.elem
                        if isinstance(resolution, And):
                            new_clauses = resolution.ands
                        else:
                            new_clauses = (resolution,)
                        novel = False
                        for new_clause in new_clauses:
                            if new_clause == bot:
                                clauses = [bot]
                                novel = True
                                break
                            elif new_clause != top and new_clause not in clauses:
                                clauses.append(new_clause)
                                novel = True
                        if novel:
                            quantifiers.extend(new_quantifiers)
                            kwargs = inner_kwargs
                            if clauses == [bot]:
                                break
                if clauses == [bot]:
                    break
        resolved = (reduce)(_coconut_pipe, [And(*clauses)] + quantifiers)
        log_simplification(self, resolved, **kwargs)
        return _coconut_tail_call(resolved.simplify, dnf=False, **kwargs)
    @_coconut_tco
    def admits_empty_universe(self):
        return _coconut_tail_call(all, (x.admits_empty_universe() for x in self.elems))
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)


_coconut_call_set_names(And)
class Eq(Expr):
    """Equality operator."""
    __slots__ = ("a", "b")
    def __init__(self, a, b):
        assert isinstance(a, Term), a
        assert isinstance(b, Term), b
        self.a, self.b = a, b
    def __repr__(self):
        return "Eq(" + repr(self.a) + ", " + repr(self.b) + ")"
    def __str__(self):
        return str(self.a) + "=" + str(self.b)
    def __eq__(self, other):
        return isinstance(other, Eq) and (self.a == other.a and self.b == other.b or self.a == other.b and self.b == other.a)
    def simplify(self, **kwargs):
        if self.a == self.b:
            return top
        else:
            return self
    @_coconut_tco
    def swap(self):
        """Swaps the order of equality."""
        return _coconut_tail_call(Eq, self.b, self.a)
    @_coconut_tco
    def find_unification(self, other):
        if isinstance(other, Quantifier):
            return _coconut_tail_call(other.find_unification, self)
        elif isinstance(other, Eq):
            a_a = self.a.find_unification(other.a)
            b_b = self.b.find_unification(other.b)
            if a_a is not None and b_b is not None:
                subs = merge_dicts(a_a, b_b)
                if subs is not None:
                    return subs
            a_b = self.a.find_unification(other.a)
            b_a = self.b.find_unification(other.b)
            if a_b is not None and b_a is not None:
                subs = merge_dicts(a_b, b_a)
                if subs is not None:
                    return subs
        else:
            return None
    @_coconut_tco
    def substitute(self, subs, **kwargs):
        return _coconut_tail_call(Eq, self.a.substitute(subs, **kwargs), self.b.substitute(subs, **kwargs))
    @_coconut_tco
    def paramodulant(self, other):
        """Create a paramodulant of other."""
        return _coconut_tail_call((sub_once), other, {self.a: self.b, self.b: self.a})
    @_coconut_tco
    def resolve_against(self, other, **kwargs):
        if isinstance(other, Not) and self.find_unification(other.neg) is not None:
            return bot
        else:
            return _coconut_tail_call(self.paramodulant, other)
    def admits_empty_universe(self):
        return self.a.admits_empty_universe() and self.b.admits_empty_universe()
    @_coconut_tco
    def __hash__(self):
# type: (...) -> int
        return _coconut_tail_call(str(self).__hash__)

    def __lt__(self, other):
# type: (...) -> int
        return str(self) < str(other)

    def __gt__(self, other):
# type: (...) -> int
        return str(self) > str(other)

    def __ge__(self, other):
# type: (...) -> int
        return str(self) >= str(other)

    def __le__(self, other):
# type: (...) -> int
        return str(self) <= str(other)

_coconut_call_set_names(Eq)
Equals = Eq



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x812bfe7c

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from pyprover.logic import FA
from pyprover.logic import TE
from pyprover.logic import top
from pyprover.logic import bot
from pyprover.logic import And
from pyprover.logic import Or
from pyprover.logic import Eq
from pyprover.tools import proves
from pyprover.tools import proves_and_proved_by
from pyprover.tools import simplify
from pyprover.tools import simplest_form
from pyprover.tools import substitute
from pyprover.tools import strict_simplify
from pyprover.tools import strict_proves
from pyprover.atoms import LowercasePropositions
from pyprover.atoms import StandardMath
from pyprover.parser import expr

# Tests:

def test_propositional_logic():
    """Runs propositional logic tests."""
    with LowercasePropositions.using(globals()):

# constructive theorems
        assert (proves)(e & f, e)
        assert (proves)(e & f, f)
        assert (proves)((e >> f, f >> g, e), g)
        assert (proves)((e >> (f >> g), e >> f, e), g)
        assert (proves)((e >> f, f >> g), e >> g)
        assert (proves)(e >> (f >> g), f >> (e >> g))
        assert (proves)(e >> (f >> g), (e >> f) >> (e >> g))
        assert (proves)(e, f >> e)
        assert (proves)(top, e >> (f >> e))
        assert (proves)(e >> (f & g), (e >> f) & (e >> g))
        assert (proves)(e >> (f >> g), (e & f) >> g)
        assert (proves)((e & f) >> g, e >> (f >> g))
        assert (proves)((e >> f) >> g, (e & f) >> g)
        assert (proves)(e & (f >> g), (e >> f) >> g)
        assert (proves)(e, e | f)
        assert (proves)(f, e | f)
        assert (proves)(e | f, f | e)
        assert (proves)(f >> g, (e | f) >> (e | g))
        assert (proves)(e, e | e)
        assert (proves)(e, ~~e)
        assert (proves)(~e, e >> f)
        assert (proves)(e >> f, ~f >> ~e)
        assert (proves)(e | f, ~(~e & ~f))
        assert (proves)(e & f, ~(~e | ~f))
        assert (proves)(~(e | f), ~e & ~f)
        assert (proves)(~e & ~f, ~(e | f))
        assert (proves)(~e | ~f, ~(e & f))
        assert (proves)(top, ~(e & ~e))
        assert (proves)(e & ~e, f)
        assert (proves)(~f, f >> g)
        assert (proves)((f >> g, ~f >> g), g)
        assert (proves)(g, f >> g)
        assert (proves)((f, ~(f & g)), ~g)
        assert (proves)((p >> r, p >> ~r), ~p)
        assert (proves)(top, (p >> ~p) >> ~p)
        assert (proves)(bot, a)
        assert (proves)(bot, top)
        assert (proves)(a, top)

# classical theorems
        assert (proves)(~~e, e)
        assert (proves)(top, e | ~e)
        assert (proves)(top, ((e >> f) >> e) >> e)
        assert (proves)(~f >> ~e, e >> f)
        assert (proves)(~(~e & ~f), e | f)
        assert (proves)(~(~e | ~f), e & f)
        assert (proves)(~(e & f), ~e | ~f)
        assert (proves)(top, (e >> f) | (f >> e))
        assert (proves)(top, (~~a | a) >> a)
        assert (proves)(p >> r, (f >> r) | (p >> g))
        assert (proves)(~(f >> g), f & ~g)
        assert (proves)(top, (~f >> bot) >> f)
        assert (proves)(f >> g, ~f | g)
        assert (proves)(top, r & s | (~r | ~s))
        assert (proves)(f >> g, g | ~f)
        assert (proves)((a & b) >> ~c, ~a | ~b | ~c)
        assert (proves)(f >> (g >> h), ~f | ~g | h)

# other theorems
        assert (proves)(t & (t >> d) | ~t & ~(t >> d), d)
        assert (proves)((f >> g, c >> d), (f | c) >> (g | d))
        assert (proves)((f >> g) >> h, f >> (g >> h))
        assert (proves)(top, p >> (s >> p))
        assert (proves)(~f | (f >> g), ~f | g)
        assert (proves)((~f, g >> f), ~g)
        assert (proves)((p >> s, r >> t, ~s | ~t), ~p | ~r)
        assert (proves)(top, ~~(f | ~f))
        assert (proves)(top, ~~(~~f >> f))
        assert (proves)((p | r, ~p), r)
        assert (proves)((f | g, ~f), g)
        assert (proves)((t | ~a, ~a | ~t), ~a)
        assert (proves)(top, ~~((~f >> bot) >> f))
        assert (proves)((s & h | ~s & ~h) & ~(s & h) | (s & ~h | ~s & h) & (s & h), ~s & ~h)
        assert (proves)(~a | ~b | ~c, (a & b) >> ~c)
        assert (proves)(~p, p >> bot)
        assert (proves)((a | b, ~a | c), b | c)
        assert (proves)(top, (f & g) >> g)
        assert (proves)((p >> s, r >> t, p | r), s | t)
        assert (proves)(top, ~p >> (p >> s))
        assert (proves)(f >> g, ((f & g) >> f) | (f >> (f & g)))
        assert (proves)((f | ~f) >> g, ~~g)
        assert (proves)(p >> r, p >> (p & r))
        assert (proves)((s & h | ~s & ~h) & (h | s) | (s & ~h | ~s & h) & ~(h | s), s & h)
        assert (proves)(~(f >> g), g >> f)
        assert (proves)((f >> g) & (f >> h), f >> (g & h))
        assert (proves)((s & h | ~s & ~h) & (~s & ~h) | (s & ~h | ~s & h) & ~(~s & ~h), ~(s & h))
        assert (proves)(top, (p >> (s >> e)) >> ((p >> s) >> (p >> e)))
        assert (proves)(top, p >> p)
        assert (proves)(~f, ~(f & g))
        assert (proves)(f >> ~f, ~f)
        assert (proves)(f >> g, ~g >> ~f)
        assert (f & g) >> f
        assert (proves)((f >> (t & a | ~t & ~a), t & ~f | ~t & f, g >> (t & ~a | ~t & a), t & g | ~t & ~g), ~a)
        assert (proves)(t & (~t & ~a) | ~t & ~(~t & ~a), a)
        assert (proves)(~f | ~g | h, f >> (g >> h))

# invalid theorems
        assert not (proves)(e >> (f >> g), (e >> f) >> g)
        assert not (proves)((e & f) >> g, (e >> f) >> g)
        assert not (proves)((e >> f) >> g, e & (f >> g))
        assert not (proves)(e, e & f)
        assert not (proves)(e | f, e & f)
        assert not (proves)(e | top, e)
        assert not (proves)(e, e & bot)
        assert not (proves)(top, bot)

# other tests
        assert And() == top
        assert Or() == bot
        assert (proves)((), top)
        assert (proves)((a ^ b, a), ~b)
        assert (proves_and_proved_by)(f >> (g >> h), (f & g) >> h)
        assert (proves_and_proved_by)(bot, a & ~a)
        assert (proves_and_proved_by)(top, a | ~a)
        assert (simplify)(top & bot) == bot
        assert simplify(a & a, b & b) == (a, b)
        assert (simplest_form)(a ^ b) == (b | a) & (~a | ~b)
        assert (simplest_form)((s & h | ~s & ~h) & ~(s & h) | (s & ~h | ~s & h) & (s & h)) == ~s & ~h
        assert (simplify)((substitute)(a ^ b, {a: True, b: False})) == top
        assert (simplify)((substitute)(a ^ b, {a: top, b: top})) == bot
        assert e << f == f >> e

def test_predicate_logic():
    """Runs predicate logic tests."""
    with StandardMath.using(globals()):

# basic tests
        assert (simplify)(FA(x, F)) == F
        assert (simplify)(TE(x, F)) == F
        assert (simplify)(FA(x, F(x)) & G) == FA(y, F(y) & G)
        assert (simplify)(FA(x, F(x)) | G) == FA(y, F(y) | G)
        assert (simplify)(TE(x, F(x)) & G) == TE(y, F(y) & G)
        assert (simplify)(TE(x, F(x)) | G) == TE(y, F(y) | G)
        assert FA(x, F(f(x))) == FA(y, F(f(y)))
        assert TE(x, F(f(x))) == TE(y, F(f(y)))
        assert TE(f, F(f(x))) == TE(g, F(g(x)))

# constructive theorems
        assert (proves)(TE(x, bot), bot)
        assert (proves)(top, FA(x, top))
        assert (proves)(FA(x, R(x) >> S(x)), FA(y, R(y)) >> FA(z, S(z)))
        assert (proves)(FA(x, R(x) & S(x)), FA(y, R(y)) & FA(z, S(z)))
        assert (proves)((FA(x, R(x) >> S(x)), TE(y, R(y))), TE(z, S(z)))
        assert (proves)(TE(x, R(x) & S(x)), TE(y, R(y)) & TE(z, S(z)))
        assert (proves)(TE(x, R(x)) | TE(y, S(y)), TE(z, R(z) | S(z)))
        assert (proves)(TE(x, R(x) | S(x)), TE(y, R(y)) | TE(z, S(z)))
        assert (proves)(FA(x, R(x)), ~TE(y, ~R(y)))
        assert (proves)(TE(x, ~R(x)), ~FA(y, R(y)))
        assert (proves)(FA(x, ~R(x)), ~TE(y, R(y)))
        assert (proves)(~TE(x, R(x)), FA(y, ~R(y)))
        assert (proves)(R(j), TE(x, R(x)))

# classical theorems
        assert (proves)(~TE(x, ~R(x)), FA(y, R(y)))
        assert (proves)(~FA(x, ~R(x)), TE(y, R(y)))
        assert (proves)(~FA(x, R(x)), TE(y, ~R(y)))
        assert (proves)(FA(x, ~~D(x)), FA(x, D(x)))
        assert (proves)(~TE(x, R(x)), FA(y, ~R(y)))
        assert (proves)(top, TE(x, D(x)) | FA(x, ~D(x)))
        assert (proves)(top, TE(x, ~D(x)) | FA(x, D(x)))
        assert (proves)(TE(x, top), TE(x, D(x) >> FA(y, D(y))))
        assert (proves)(TE(x, ~~D(x)), TE(x, D(x)))
        assert (proves)(FA(x, C(x) | D(x)), FA(x, C(x)) | TE(x, D(x)))

# other theorems
        assert (proves)(FA(x, H(j) >> T(x)), H(j) >> FA(x, T(x)))
        assert (proves)(TE(x, R(x) >> B(x)), FA(x, R(x)) >> TE(x, B(x)))
        assert (proves)(~FA(x, bot), TE(x, top))
        assert (proves)(FA(x, TE(y, F(y) | G(x))), FA(x, G(x) | TE(x, F(x))))
        assert (proves)((FA(x, FA(y, FA(z, (S(x, y) & S(y, z)) >> S(x, z)))), ~TE(x, S(x, x))), FA(x, FA(y, S(x, y) >> ~S(y, x))))
        assert (proves)(FA(x, G(x)) | FA(x, B(x)), FA(x, G(x) | B(x)))
        assert (proves)(TE(z, FA(k, P(z, k))), FA(y, TE(x, P(x, y))))
        assert (proves)(TE(x, C(x) & B(x)), TE(x, B(x) & C(x)))
        assert (proves)(TE(x, C(x, i) & B(x, j)), TE(x, C(x, i) >> B(x, j)))
        assert (proves)(FA(x, C(x) & B(x)), FA(x, B(x) & C(x)))
        assert (proves)(FA(x, C(x) & B(x)), FA(x, C(x)) & FA(x, B(x)))
        assert (proves)(FA(x, bot), ~TE(x, top))
        assert (proves)((~TE(x, G(x)) | FA(x, F(x)), C(j) >> FA(x, D(x))), FA(y, FA(z, ~G(z) | F(y) & C(j) >> D(y))))
        assert (proves)(FA(x, G(x)) | TE(x, F(x)), FA(x, TE(y, F(y) | G(x))))
        assert (proves)((P | TE(x, W)) >> FA(z, R), FA(z, FA(x, (P | W) >> R)))
        assert (proves)((TE(x, ~F(x)), FA(x, F(x))), bot)
        assert (proves)((FA(x, ~F(x)), FA(x, F(x))), FA(x, bot))
        assert (proves)(TE(x, F(x)) | TE(x, G(x)), TE(x, TE(y, F(x) | G(y))))
        assert (proves)(TE(x, FA(y, P(x, y))), FA(y, TE(x, P(x, y))))
        assert (proves)(TE(x, FA(y, bot)), bot)
        assert (proves)(top, FA(x, TE(y, top)))
        assert (proves)(FA(x, TE(y, F(y)) | G(x)), FA(x, TE(y, F(y) | G(x))))
        assert (proves)(TE(x, FA(y, F(y) & G(x))), TE(x, FA(y, F(y)) & G(x)))
        assert (proves)(TE(x, ~R(x)), TE(y, R(y) >> (R(j) & R(k))))
        assert (proves)(P(c), TE(x, P(x)))
        assert (proves)(P(c), TE(x, top))
        assert (proves)(P(c) & ~P(c), TE(x, top))
        assert (proves)(P(c) | ~P(c), TE(x, top))

# invalid theorems
        assert not (proves)(FA(x, R(x)) >> FA(y, S(y)), FA(z, R(z) >> S(z)))
        assert not (proves)(TE(x, R(x)) & TE(y, S(y)), TE(z, R(z) & S(z)))
        assert not (proves)(TE(x, R(x)), FA(y, R(y)))

# non-empty universe theorems
        assert (proves)(top, TE(x, top))
        assert (proves)(top, TE(x, D(x) >> FA(y, D(y))))
        assert (proves)((R(j), FA(x, R(x) >> S(x))), S(j))
        assert (proves)(FA(x, R(x)) >> FA(y, S(y)), TE(x, FA(y, ~R(x) | S(y))))
        assert (proves)(FA(x, R(x)), TE(y, R(y)))
        assert (proves)((T(i), FA(x, T(x) >> T(s(x)))), T(s(i)))
        assert (proves)(top, TE(x, R(x) >> (R(j) & R(k))))
        assert (proves)((FA(x, ~F(x)), FA(x, F(x))), bot)

# equality theorems
        assert (proves)(top, Eq(a, a))
        assert (proves)(Eq(a, b) & Eq(b, c), Eq(a, c))
        assert (proves)(Eq(a, b) & Eq(b, c), Eq(c, a))
        assert (proves)(Eq(a, b) & F(a), F(b))
        assert (proves)((Eq(a, b) | Eq(a, c), F(a)), F(b) | F(c))
        assert (proves)(FA(x, Eq(a, x)), Eq(a, b))
        assert (proves)(Eq(a, b), Eq(b, a))
        assert (proves)(Eq(a, b), Eq(f(a), f(b)))

def test_empty_universe():
    """Runs predicate logic tests in a potentially empty universe."""
    with StandardMath.using(globals()):

# basic tests
        assert (strict_simplify)(FA(x, F)) != F
        assert (strict_simplify)(TE(x, F)) != F
        assert (strict_simplify)(FA(x, F(x)) & G) != FA(y, F(y) & G)
        assert (strict_simplify)(FA(x, F(x)) | G) == FA(y, F(y) | G)
        assert (strict_simplify)(TE(x, F(x)) & G) == TE(y, F(y) & G)
        assert (strict_simplify)(TE(x, F(x)) | G) != TE(y, F(y) | G)

# constructive theorems
        assert (strict_proves)(TE(x, bot), bot)
        assert (strict_proves)(top, FA(x, top))
        assert (strict_proves)(FA(x, R(x) >> S(x)), FA(y, R(y)) >> FA(z, S(z)))
        assert (strict_proves)(FA(x, R(x) & S(x)), FA(y, R(y)) & FA(z, S(z)))
        assert (strict_proves)((FA(x, R(x) >> S(x)), TE(y, R(y))), TE(z, S(z)))
        assert (strict_proves)(TE(x, R(x) & S(x)), TE(y, R(y)) & TE(z, S(z)))
        assert (strict_proves)(TE(x, R(x)) | TE(y, S(y)), TE(z, R(z) | S(z)))
        assert (strict_proves)(TE(x, R(x) | S(x)), TE(y, R(y)) | TE(z, S(z)))
        assert (strict_proves)(FA(x, R(x)), ~TE(y, ~R(y)))
        assert (strict_proves)(TE(x, ~R(x)), ~FA(y, R(y)))
        assert (strict_proves)(FA(x, ~R(x)), ~TE(y, R(y)))
        assert (strict_proves)(~TE(x, R(x)), FA(y, ~R(y)))
        assert (strict_proves)(R(j), TE(x, R(x)))

# classical theorems
        assert (strict_proves)(~TE(x, ~R(x)), FA(y, R(y)))
        assert (strict_proves)(~FA(x, ~R(x)), TE(y, R(y)))
        assert (strict_proves)(~FA(x, R(x)), TE(y, ~R(y)))
        assert (strict_proves)(FA(x, ~~D(x)), FA(x, D(x)))
        assert (strict_proves)(~TE(x, R(x)), FA(y, ~R(y)))
        assert (strict_proves)(top, TE(x, D(x)) | FA(x, ~D(x)))
        assert (strict_proves)(top, TE(x, ~D(x)) | FA(x, D(x)))
        assert (strict_proves)(TE(x, top), TE(x, D(x) >> FA(y, D(y))))
        assert (strict_proves)(TE(x, ~~D(x)), TE(x, D(x)))
        assert (strict_proves)(FA(x, C(x) | D(x)), FA(x, C(x)) | TE(x, D(x)))

# other theorems
        assert (strict_proves)(FA(x, H(j) >> T(x)), H(j) >> FA(x, T(x)))
        assert (strict_proves)(TE(x, R(x) >> B(x)), FA(x, R(x)) >> TE(x, B(x)))
        assert (strict_proves)(~FA(x, bot), TE(x, top))
        assert (strict_proves)(FA(x, TE(y, F(y) | G(x))), FA(x, G(x) | TE(x, F(x))))
        assert (strict_proves)((FA(x, FA(y, FA(z, (S(x, y) & S(y, z)) >> S(x, z)))), ~TE(x, S(x, x))), FA(x, FA(y, S(x, y) >> ~S(y, x))))
        assert (strict_proves)(FA(x, G(x)) | FA(x, B(x)), FA(x, G(x) | B(x)))
        assert (strict_proves)(TE(z, FA(k, P(z, k))), FA(y, TE(x, P(x, y))))
        assert (strict_proves)(TE(x, C(x) & B(x)), TE(x, B(x) & C(x)))
        assert (strict_proves)(TE(x, C(x, i) & B(x, j)), TE(x, C(x, i) >> B(x, j)))
        assert (strict_proves)(FA(x, C(x) & B(x)), FA(x, B(x) & C(x)))
        assert (strict_proves)(FA(x, C(x) & B(x)), FA(x, C(x)) & FA(x, B(x)))
        assert (strict_proves)(FA(x, bot), ~TE(x, top))
        assert (strict_proves)((~TE(x, G(x)) | FA(x, F(x)), C(j) >> FA(x, D(x))), FA(y, FA(z, ~G(z) | F(y) & C(j) >> D(y))))
        assert (strict_proves)(FA(x, G(x)) | TE(x, F(x)), FA(x, TE(y, F(y) | G(x))))
        assert (strict_proves)((P | TE(x, W)) >> FA(z, R), FA(z, FA(x, (P | W) >> R)))
        assert (proves)(TE(x, F(x)) | TE(x, G(x)), TE(x, TE(y, F(x) | G(y))))
        assert (proves)(TE(x, FA(y, P(x, y))), FA(y, TE(x, P(x, y))))
        assert (proves)(TE(x, FA(y, bot)), bot)
        assert (proves)(top, FA(x, TE(y, top)))
        assert (proves)(FA(x, TE(y, F(y)) | G(x)), FA(x, TE(y, F(y) | G(x))))
        assert (proves)(TE(x, FA(y, F(y) & G(x))), TE(x, FA(y, F(y)) & G(x)))
        assert (strict_proves)(TE(x, ~R(x)), TE(y, R(y) >> (R(j) & R(k))))
        assert (strict_proves)(P(c), TE(x, P(x)))
        assert (strict_proves)(P(c), TE(x, top))
        assert (strict_proves)(P(c) & ~P(c), TE(x, top))
        assert (strict_proves)(P(c) | ~P(c), TE(x, top))

# invalid theorems
        assert not (strict_proves)(FA(x, R(x)) >> FA(y, S(y)), FA(z, R(z) >> S(z)))
        assert not (strict_proves)(TE(x, R(x)) & TE(y, S(y)), TE(z, R(z) & S(z)))
        assert not (strict_proves)(TE(x, R(x)), FA(y, R(y)))

# non-empty universe theorems
        assert not (strict_proves)(top, TE(x, top))
        assert not (strict_proves)(top, TE(x, D(x) >> FA(y, D(y))))
        assert not (strict_proves)((R(j), FA(x, R(x) >> S(x))), S(j))
        assert not (strict_proves)(FA(x, R(x)) >> FA(y, S(y)), TE(x, FA(y, ~R(x) | S(y))))
        assert not (strict_proves)(FA(x, R(x)), TE(y, R(y)))
        assert not (strict_proves)((T(i), FA(x, T(x) >> T(s(x)))), T(s(i)))
        assert not (strict_proves)(top, TE(x, R(x) >> (R(j) & R(k))))
        assert not (strict_proves)((FA(x, ~F(x)), FA(x, F(x))), bot)

# equality theorems
        assert (strict_proves)(top, Eq(a, a))
        assert (strict_proves)(Eq(a, b) & Eq(b, c), Eq(a, c))
        assert (strict_proves)(Eq(a, b) & Eq(b, c), Eq(c, a))
        assert (strict_proves)(Eq(a, b) & F(a), F(b))
        assert (strict_proves)((Eq(a, b) | Eq(a, c), F(a)), F(b) | F(c))
        assert (strict_proves)(FA(x, Eq(a, x)), Eq(a, b))
        assert (strict_proves)(Eq(a, b), Eq(b, a))
        assert (strict_proves)(Eq(a, b), Eq(f(a), f(b)))

def test_parser():
    """Tests math notation parsing."""
    with StandardMath.using(globals()):
        assert expr(r"A") == A
        assert expr(r"F(x)") == F(x)
        assert expr(r"F(f(x))") == F(f(x))
        assert expr(r"A x. F(x)") == FA(x, F(x))
        assert expr(r"A x. E y. F(x) \/ G(y)") == FA(x, TE(y, F(x) | G(y)))
        assert expr(r"F(i) /\ A x. G(x)") == F(i) & FA(x, G(x))
        assert expr(r"F -> G -> H") == F >> (G >> H)
        assert expr(r"~F(x)") == ~F(x)
        assert expr(r"-G(x) /\ F(x)") == ~G(x) & F(x)
        assert expr(r"A /\ (B \/ C)") == A & (B | C)
        assert expr(r"a = b") == Eq(a, b)
        assert expr(r"forall x: A, B(x)") == FA(x, A(x) >> B(x))
        assert expr(r"exists x: A, B(x)") == TE(x, A(x) & B(x))

if __name__ == "__main__":
    test_propositional_logic()
    test_predicate_logic()
    test_empty_universe()
    test_parser()
    print("<success>")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# __coconut_hash__ = 0x78fad73

# Compiled with Coconut version 1.5.0-post_dev58 [Fish License]

# Coconut Header: -------------------------------------------------------------

from __future__ import print_function, absolute_import, unicode_literals, division
import sys as _coconut_sys, os as _coconut_os
_coconut_file_dir = _coconut_os.path.dirname(_coconut_os.path.abspath(__file__))
_coconut_cached_module = _coconut_sys.modules.get(str("__coconut__"))
if _coconut_cached_module is not None and _coconut_os.path.dirname(_coconut_cached_module.__file__) != _coconut_file_dir:
    del _coconut_sys.modules[str("__coconut__")]
_coconut_sys.path.insert(0, _coconut_file_dir)
_coconut_module_name = _coconut_os.path.splitext(_coconut_os.path.basename(_coconut_file_dir))[0]
if _coconut_module_name and _coconut_module_name[0].isalpha() and all(c.isalpha() or c.isdigit() for c in _coconut_module_name) and "__init__.py" in _coconut_os.listdir(_coconut_file_dir):
    _coconut_full_module_name = str(_coconut_module_name + ".__coconut__")
    import __coconut__ as _coconut__coconut__
    _coconut__coconut__.__name__ = _coconut_full_module_name
    for _coconut_v in vars(_coconut__coconut__).values():
        if getattr(_coconut_v, "__module__", None) == str("__coconut__"):
            try:
                _coconut_v.__module__ = _coconut_full_module_name
            except AttributeError:
                type(_coconut_v).__module__ = _coconut_full_module_name
    _coconut_sys.modules[_coconut_full_module_name] = _coconut__coconut__
from __coconut__ import *
from __coconut__ import _coconut_tail_call, _coconut_tco, _coconut_call_set_names, _coconut, _coconut_MatchError, _coconut_igetitem, _coconut_base_compose, _coconut_forward_compose, _coconut_back_compose, _coconut_forward_star_compose, _coconut_back_star_compose, _coconut_forward_dubstar_compose, _coconut_back_dubstar_compose, _coconut_pipe, _coconut_star_pipe, _coconut_dubstar_pipe, _coconut_back_pipe, _coconut_back_star_pipe, _coconut_back_dubstar_pipe, _coconut_none_pipe, _coconut_none_star_pipe, _coconut_none_dubstar_pipe, _coconut_bool_and, _coconut_bool_or, _coconut_none_coalesce, _coconut_minus, _coconut_map, _coconut_partial, _coconut_get_function_match_error, _coconut_base_pattern_func, _coconut_addpattern, _coconut_sentinel, _coconut_assert, _coconut_mark_as_match, _coconut_reiterable
_coconut_sys.path.pop(0)
# Compiled Coconut: -----------------------------------------------------------

# Imports:

from pyprover.logic import ForAll
from pyprover.logic import Exists
from pyprover.logic import FA
from pyprover.logic import TE
from pyprover.logic import top
from pyprover.logic import bot
from pyprover.logic import true
from pyprover.logic import false
from pyprover.logic import Not
from pyprover.logic import Imp
from pyprover.logic import Or
from pyprover.logic import And
from pyprover.logic import Eq
from pyprover.tools import props
from pyprover.tools import terms
from pyprover.tools import solve
from pyprover.tools import strict_solve
from pyprover.tools import no_proof_of
from pyprover.tools import proves
from pyprover.tools import strict_proves
from pyprover.tools import proves_and_proved_by
from pyprover.tools import strict_proves_and_proved_by
from pyprover.tools import iff
from pyprover.tools import simplify
from pyprover.tools import strict_simplify
from pyprover.tools import simplest_form
from pyprover.tools import strict_simplest_form
from pyprover.tools import simplest_solution
from pyprover.tools import strict_simplest_solution
from pyprover.tools import substitute
from pyprover.atoms import LowercasePropositions
from pyprover.atoms import UppercasePropositions
from pyprover.atoms import LowercaseVariables
from pyprover.atoms import UppercaseVariables
from pyprover.atoms import StandardMath
from pyprover.parser import expr

# Main:

StandardMath.use(globals())
